from typing import Any, Callable, Union

import pint

from .util import ensure_units
from ._generator import UnitGenerator


def to_units(
    units: Union[pint.Unit, UnitGenerator]
) -> Callable[[Any], pint.Quantity]:
    """
    Create a callable ``f(x)`` returning
    :func:`ensure_units(x, units, convert=False) <ensure_units>`.

    :param units:
        Units to ensure conversion to.
    """

    def f(x):
        return ensure_units(x, units)

    return f



import pint

#: Default unit registry
unit_registry = pint.UnitRegistry()


def set_unit_registry(ureg: pint.UnitRegistry) -> None:
    """
    Set unit registry. By default, Pinttrs has its own registry.

    :param ureg: Unit registry.
    :raises: :class:`TypeError` if ``ureg`` is not a :class:`pint.UnitRegistry`.
    """
    global unit_registry
    if not isinstance(ureg, pint.UnitRegistry):
        raise TypeError("ureg must be a pint.UnitRegistry")
    unit_registry = ureg


def get_unit_registry() -> pint.UnitRegistry:
    """
    Get default unit registry.
    """
    global unit_registry
    return unit_registry



from contextlib import contextmanager
from copy import copy
from typing import Union, Callable

import attr
import pint


@attr.s
class UnitGenerator:
    """
    A callable object which returns units objects.
    Stored units can be contextually overridden using the :meth:`.override`
    method.

    .. seealso:: :class:`.UnitContext`

    :Attributes / constructor arguments:

        **units** (:class:`pint.Unit` or :class:`Callable`) –
        Stored units or generator.
    """

    units: Union[pint.Unit, Callable] = attr.ib()

    def __call__(self) -> pint.Unit:
        """
        :returns:
            If ``units`` is a :class:`pint.Unit`, it is returned; if ``units``
            is a callable (typically, another :class:`UnitGenerator`), the
            result of its evaluation will be returned.
        """
        if callable(self.units):
            return self.units()
        return self.units

    @contextmanager
    def override(self, units: Union[pint.Unit, Callable, str]) -> None:
        """
        Temporarily override the value of ``units``. The initial value of
        ``units`` is restored upon leaving context.

        :param units:
            Temporary replacement for ``units``. String values are interpreted
            based on the unit registry of currently stored units.
        """
        units_old = copy(self.units)

        if isinstance(units, str):  # Safeguard to convert strings
            if callable(self.units):
                self.units = self.units()._REGISTRY.Unit(units)
            else:
                self.units = self.units._REGISTRY.Unit(units)
        else:
            self.units = units
        try:
            yield
        finally:
            self.units = units_old



from typing import Union

import attr
import pint
from attr import NOTHING

from ._generator import UnitGenerator
from ._metadata import MetadataKey
from .converters import to_units
from .validators import has_compatible_units


def attrib(
    default=NOTHING,
    validator=NOTHING,
    repr=NOTHING,
    cmp=None,
    hash=None,
    init=True,
    metadata=None,
    type=None,
    converter=NOTHING,
    factory=None,
    kw_only=False,
    eq=None,
    order=None,
    on_setattr=NOTHING,
    units: Union[None, pint.Unit, UnitGenerator] = None,
):
    """
    Create a new attribute on a class, possibly with units. This function
    wraps :func:`attr.ib` and retains its behaviour unless otherwise specified.

    :param validator:
        If set to :class:`~attr.NOTHING` and ``units`` is not ``None``, defaults
        to :func:`~pinttr.validators.has_compatible_units` (possibly wrapped in
        :func:`attr.validators.optional` if ``default`` is ``None``).
        Otherwise retains original behaviour.

    :param repr:
        If set to :class:`~attr.NOTHING` and ``units`` is not ``None``, defaults
        to a callable printing quantities nicely.
        Otherwise retains original behaviour.

    :param converter:
        If set to :class:`~attr.NOTHING` and ``units`` is not ``None``, defaults
        to :func:`to_units(units) <pinttr.converters.to_units>`
        (possibly wrapped in :func:`attr.converters.optional` if ``default`` is
        ``None``). Otherwise retains original behaviour.

    :param on_setattr:
        If set to :class:`~attr.NOTHING` and ``units`` is not ``None``,
        defaults to
        ``attr.setters.pipe(attr.setters.convert, attr.setters.validate)``.
        Otherwise retains original behaviour.

    :param units:
        Default units attached to the defined attribute. Accepts a
        :class:`UnitGenerator` instance. Has no effect if set to ``None``.

    .. versionchanged:: 21.3.0
       Added prettier default repr.
    """

    # Initialise attr.ib arguments
    metadata = dict() if not metadata else metadata

    # Process declared compatible units
    if units is not None:
        # Set field metadata
        if isinstance(units, UnitGenerator):
            unit_generator = units

        elif isinstance(units, pint.Unit):
            unit_generator = UnitGenerator(units)

        else:
            raise TypeError("Argument 'units' must be a pint.Units or a UnitGenerator")

        metadata[MetadataKey.UNITS] = unit_generator

        # Set field converter
        if converter is NOTHING:
            if default is None:
                converter = attr.converters.optional(to_units(unit_generator))
            else:
                converter = to_units(unit_generator)

        # Set field validator
        if validator is NOTHING:
            if default is None:
                validator = attr.validators.optional(has_compatible_units)
            else:
                validator = has_compatible_units

        # Ensure that unit conversion and validation is carried out upon setting
        if on_setattr is NOTHING:
            on_setattr = attr.setters.pipe(attr.setters.convert, attr.setters.validate)

        # Set field repr
        if repr is NOTHING:

            def f(x):
                if isinstance(x, pint.Quantity):
                    return f"{x:~P}"
                else:
                    return x.__repr__()

            repr = f

    # If one of the following hasn't been set because units is unset, we set it
    # to the original default value
    if converter is NOTHING:
        converter = None
    if validator is NOTHING:
        validator = None
    if on_setattr is NOTHING:
        on_setattr = None
    if repr is NOTHING:
        repr = True

    return attr.ib(
        default=default,
        validator=validator,
        repr=repr,
        cmp=cmp,
        hash=hash,
        init=init,
        metadata=metadata,
        type=type,
        converter=converter,
        factory=factory,
        kw_only=kw_only,
        eq=eq,
        order=order,
        on_setattr=on_setattr,
    )



import enum


class MetadataKey(enum.Enum):
    """Attribute metadata keys.

    These Enum values should be used as metadata attribute keys. They are
    immutable and guarantee that no collision can occur with another piece of
    code.
    """

    # Units compatible with this field (callable)
    UNITS = 0



from . import converters, exceptions, util, validators
from ._context import UnitContext
from ._defaults import get_unit_registry, set_unit_registry
from ._generator import UnitGenerator
from ._interpret import interpret_units
from ._make import attrib
from ._next_gen import field

# Package metadata
__version__ = "21.3.1"

# Other definitions
ib = attrib
__all__ = [
    "UnitContext",
    "UnitGenerator",
    "attrib",
    "converters",
    "exceptions",
    "field",
    "get_unit_registry",
    "ib",
    "interpret_units",
    "set_unit_registry",
    "util",
    "validators",
]



from ._metadata import MetadataKey
from .exceptions import UnitsError
from .util import units_compatible


def has_compatible_units(instance, attribute, value):
    """
    Validate if ``value`` has units compatible (in the sense of
    :func:`.units_compatible`) with ``attribute``.
    Only works with unit-enabled fields created with :func:`pinttr.ib`.
    """

    compatible_units = attribute.metadata[MetadataKey.UNITS]()

    try:
        if not units_compatible(value.units, compatible_units):
            raise UnitsError(
                units1=value.units,
                units2=compatible_units,
                extra_msg=f": incompatible units '{value.units}' "
                f"used to set field '{attribute.name}' "
                f"(allowed: '{compatible_units}').",
            )

    except AttributeError:  # value.units doesn't exist
        raise UnitsError(
            units1=None,
            units2=compatible_units,
            extra_msg=f": unitless value '{value}' "
            f"used to set field '{attribute.name}' "
            f"(requires units '{compatible_units}').",
        )



from pint import DimensionalityError


class UnitsError(DimensionalityError):
    """
    Raised when encountering issues with units (can be raised even when
    :class:`.DimensionalityError` would not).
    """



from typing import Any, Callable, Union

import pint


def always_iterable(obj, base_type=(str, bytes)):
    """
    Ensure that the object it is passed is iterable.

    - If ``obj`` is iterable, return an iterator over its items.
    - If ``obj`` is not iterable, return a one-item iterable containing ``obj``.
    - If ``obj`` is ``None``, return an empty iterable.

    By default, binary and text strings are not considered iterable.
    If ``base_type`` is set, objects for which ``isinstance(obj, base_type)``
    returns ``True`` won't be considered iterable.

    Set ``base_type`` to ``None`` to avoid any special handling and treat
    objects Python considers iterable as iterable.

    .. note::

        Copied from :func:`more_itertools.always_iterable`.
    """
    if obj is None:
        return iter(())

    if (base_type is not None) and isinstance(obj, base_type):
        return iter((obj,))

    try:
        return iter(obj)
    except TypeError:
        return iter((obj,))


def ensure_units(
    value: Any,
    default_units: Union[pint.Unit, Callable],
    convert: bool = False,
) -> pint.Quantity:
    """Ensure that a value is wrapped in a Pint quantity container.

    :param value:
        Value to ensure the wrapping of.

    :param default_units:
        Units to use to initialise the :class:`pint.Quantity` if ``value`` is
        not a :class:`pint.Quantity`. A callable can be passed;
        in this case, the applied units will be ``default_units()``.

    :param convert:
        If ``True``, ``value`` will also be converted to ``default_units`` if it
        is a :class:`pint.Quantity`.

    :returns:
        Converted ``value``.

    .. versionchanged:: 21.1.0
       Relocated from ``pinttr.converters`` to ``pinttr.util``.
    """
    if callable(default_units):
        units = default_units()
    else:
        units = default_units

    if not isinstance(units, pint.Unit):
        raise TypeError("Argument 'units' must be a pint.Units or a UnitGenerator")

    if isinstance(value, pint.Quantity):
        if convert:
            return value.to(units)
        else:
            return value
    else:
        return value * units


def units_compatible(unit1: pint.Unit, unit2: pint.Unit) -> bool:
    """
    Check if two units are compatible. Accounts for angle units.

    :param unit1:
        First unit to check for compatibility.

    :param unit2:
        Second unit to check for compatibility.

    :returns:
        ``True`` if ``unit1`` and ``unit2`` have the same dimensionality,
        ``False`` otherwise.
    """
    return (1.0 * unit1 / unit2).unitless



from copy import copy
from typing import Any, Dict, Union

import pint

from ._defaults import get_unit_registry


def interpret_units(
    d: Dict[str, Any],
    ureg: Union[pint.UnitRegistry, None] = None,
    inplace: bool = False,
) -> Dict[str, Any]:
    """
    Interpret units in a dictionary. The dictionary is searched for matching
    magnitude-units field pairs. For a magnitude field with key ``"x"``, the
    corresponding unit field is ``"x_units"``. For each pair found, the
    magnitude field is attached units and converted to a :class:`pint.Quantity`
    object. The unit field is then dropped.

    If the magnitude field is already a Pint quantity, it will be converted to
    specified units (and conversion will fail if units are incompatible).

    .. admonition:: Example

       .. code-block:: python

          {
              "field": 1.0,
              "field_units": "m"
          }

       will be interpreted as

       .. code-block:: python

          {"field": ureg.Quantity(1.0, "m")}

    .. warning:: Dictionary keys must be strings.

    :param d:
        Dictionary in which units will be interpreted.

    :param ureg:
        Unit registry to use for unit creation. If set to ``None``,
        Pinttrs's registered unit registry is used. See also
        :func:`pinttr.set_unit_registry`.

    :param inplace:
        If ``True``, modify the dictionary in-place; otherwise,
        return a modified copy.

    :returns:
        A copy of ``d``, where unit fields are interpreted using ``ureg``
        to attach units to the corresponding magnitude field.

    .. versionchanged:: 1.1.0
       Support for converting quantity magnitude fields.
    """
    if ureg is None:
        ureg = get_unit_registry()

    if not inplace:
        result = copy(d)
    else:
        result = d

    for key in list(result.keys()):
        if key.endswith("_units"):
            magnitude_key = key[:-6]

            try:
                magnitude = result[magnitude_key]
            except KeyError:
                continue

            units = result[key]

            # If magnitude value is a quantity, convert to requested units
            # (and thus check for unit compatibility)
            if isinstance(magnitude, pint.Quantity):
                magnitude = magnitude.m_as(units)

            result[magnitude_key] = ureg.Quantity(magnitude, result[key])
            del result[key]

    return result



from contextlib import ExitStack, contextmanager
from typing import Callable, Dict, Hashable, Optional, Union

import attr
import pint

from ._defaults import get_unit_registry
from ._func import identity
from ._generator import UnitGenerator


@attr.s
class UnitContext:
    """
    An overridable registry of :class:`.UnitGenerator` objects.

    This class maintains a registry of :class:`.UnitGenerator` instances.
    Stored :class:`.UnitGenerator` objects can be conveniently overridden using
    the :meth:`.override` context manager.

    :Attributes / constructor arguments:

        * **registry** (Dict[Hashable, :class:`.UnitGenerator`]) –
          Unit generator registry. Keys can be any hashable type, but
          :class:`str` or :class:`~enum.Enum` are recommended.
          Defaults to an empty dictionary.

          .. note:: The initialisation sequence will make repeated calls to
             :meth:`register` and will consequently apply the same key and value
             conversion rules.

        * **interpret_str** (:class:`bool`) –
          If ``True``, attempt string-to-units interpretation when specifying
          unit generators as :class:`str`.

        * **ureg** (Optional[:class:`~pint.UnitRegistry`]) –
          Unit registry used for string-to-units interpretation. If ``None``,
          the default registry is used (see :func:`.get_unit_registry`).

        * **key_converter** (Callable) –
          Converter used for keys. Defaults to :func:`.identity`.

    .. versionchanged:: 1.1.0
       Added ``ureg``.
    """

    registry: Dict[Hashable, UnitGenerator] = attr.ib(factory=dict)
    interpret_str: bool = attr.ib(default=False)
    ureg: Optional[pint.UnitRegistry] = attr.ib(default=None)
    key_converter: Callable = attr.ib(default=identity)

    def __attrs_post_init__(self):
        # Convert keys when relevant
        for key in list(self.registry.keys()):
            self._convert_key(key)
        # Convert values when relevant
        for key in self.registry.keys():
            self._convert_value(key)

    def _convert_key(self, key):
        """
        Apply in-place conversion rule ``key_converter`` to a registered key.

        :param key:
            Key to which conversion is to be applied.
        """
        self.registry[self.key_converter(key)] = self.registry.pop(key)

    def _convert_value(self, key):
        """
        Apply conversion rules to a registered value. Registry values specified
        as :class:`pint.Unit` will be converted to :class:`UnitGenerator`
        instances. If string-to-units interpretation is activated, units will be
        converted to :class:`pint.Unit` objects using ``self.ureg`` if it is
        set, or the default registry returned by :func:`.get_unit_registry`
        otherwise.

        :param key:
            Key to the value to which conversion is to be applied.
        """
        key = self.key_converter(key)
        value = self.registry[key]

        # Interpret units specified as string if necessary
        if isinstance(value, str):
            if self.interpret_str:
                value = (
                    self.ureg.Unit(value)
                    if self.ureg is not None
                    else get_unit_registry().Unit(value)
                )
            else:
                raise TypeError("String-to-units interpretation is disabled")

        # Proceed with actual registration
        if isinstance(value, pint.Unit):
            self.registry[key] = UnitGenerator(value)
        elif isinstance(value, UnitGenerator):
            self.registry[key] = value
        else:
            raise TypeError(
                f"Items must be either str, pint.Unit or UnitGenerator; "
                f"found: {key}: {type(value)}"
            )

    def register(
        self, key: Hashable, value: Union[UnitGenerator, pint.Unit, str]
    ) -> None:
        """
        Add or update an entry in the registry. Conversion rules are applied as
        follows:

        * ``key`` is applied the ``key_converter`` converter;
        * ``value`` is converted to a :class:`UnitGenerator`.

        In addition, if ``interpret_str`` is ``True``, ``value`` can be
        specified as a string. In that case, it will be converted to a
        :class:`pint.Unit` using the unit registry returned by
        :func:`.get_unit_registry`.

        :param key:
            Key to the registered entry.

        :param value:
            Object to register.
        """
        self.registry[key] = value
        self._convert_key(key)
        self._convert_value(key)

    def update(self, d: Dict) -> None:
        """
        Update the registry with a dictionary.

        :param d:
            Dictionary used to apply :meth:`register` for each of its key-value
            pairs.
        """
        for key, value in d.items():
            self.register(key, value)

    def get(self, key: Hashable) -> pint.Unit:
        """
        Evaluate :class:`UnitGenerator` instance registered as ``key``.

        :param key:
            Key to the :class:`UnitGenerator` to evaluate. The ``key_converter``
            is applied.

        :returns:
            Evaluated units.
        """
        key = self.key_converter(key)

        try:
            return self.registry[key]()
        except KeyError:
            raise

    def get_all(self) -> Dict[Hashable, pint.Unit]:
        """
        Evaluate all registered :class:`UnitGenerator` instance.

        :returns:
            Evaluated units as a dictionary.
        """
        return {key: self.get(key) for key in self.registry.keys()}

    def deferred(self, key: Hashable) -> UnitGenerator:
        """
        Return the :class:`UnitGenerator` registered with a given key.

        :param key:
            Key to the :class:`UnitGenerator` to return. The ``key_converter``
            is applied.

        :returns:
            Unit generator.
        """
        key = self.key_converter(key)
        return self.registry[key]

    @contextmanager
    def override(self, *args, **kwargs) -> None:
        """
        Temporarily override underlying unit generators. This method acts as a
        convenience proxy for :meth:`UnitGenerator.override`.

        Override specifications can take multiple forms:

        * an arbitrary number of dictionaries can be passed as positional
          arguments;
        * key-value pairs may also be specified as keyword arguments.

        Both approaches can be mixed.

        .. note:: When using the keyword argument specification, passed values
           will systematically be strings. Consequently, either

           * registry keys must be strings;
           * or the ``key_converter`` must provide the conversion protocol for
             string-valued keys.
        """
        with ExitStack() as stack:
            for arg in args:
                if not isinstance(arg, dict):
                    raise TypeError
                for key, value in arg.items():
                    stack.enter_context(
                        self.registry[self.key_converter(key)].override(value)
                    )

            for key, value in kwargs.items():
                stack.enter_context(
                    self.registry[self.key_converter(key)].override(value)
                )

            try:
                yield
            finally:
                pass

    def __getitem__(self, key: Hashable) -> pint.Unit:
        """
        Alias to :meth:`.get`.

        :param key:
            Key to the :class:`UnitGenerator` to evaluate. The ``key_converter``
            is applied.

        :returns:
            Evaluated units.

        .. versionadded:: 21.2.0
        """
        return self.get(key)

    def __setitem__(self, key: Hashable, value: Union[UnitGenerator, pint.Unit, str]) -> None:
        """
        Alias to :meth:`register`.

        :param key:
            Key to the registered entry.

        :param value:
            Object to register.

        .. versionadded:: 21.2.0
        """
        self.register(key, value)


"""
APIs in the style of attrs's next-generation APIs.
"""
from typing import Union

import pint
from attr import NOTHING

from ._generator import UnitGenerator
from ._make import attrib


def field(
    *,
    default=NOTHING,
    validator=NOTHING,
    repr=NOTHING,
    hash=None,
    init=True,
    metadata=None,
    converter=NOTHING,
    factory=None,
    kw_only=False,
    eq=None,
    order=None,
    on_setattr=NOTHING,
    units: Union[None, pint.Unit, UnitGenerator] = None,
):
    """
    Identical to :func:`pinttr.ib`, except keyword-only and with some arguments
    removed.

    .. versionadded:: 21.3.0
    """
    return attrib(
        default=default,
        validator=validator,
        repr=repr,
        hash=hash,
        init=init,
        metadata=metadata,
        converter=converter,
        factory=factory,
        kw_only=kw_only,
        eq=eq,
        order=order,
        on_setattr=on_setattr,
        units=units,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pinttr']

package_data = \
{'': ['*']}

install_requires = \
['attrs', 'pint>=0.16']

setup_kwargs = {
    'name': 'pinttrs',
    'version': '21.3.1',
    'description': 'Pint meets attrs',
    'long_description': "# Pinttrs\n\n*Pint meets attrs*\n\n[![PyPI version](https://img.shields.io/pypi/v/pinttrs?color=blue&style=flat-square)](https://pypi.org/project/pinttrs)\n[![Conda version](https://img.shields.io/conda/v/eradiate/pinttrs?color=blue&style=flat-square)](https://anaconda.org/eradiate/pinttrs)\n\n[![GitHub Workflow Status (branch)](https://img.shields.io/github/workflow/status/leroyvn/pinttrs/Tests/main?style=flat-square)](https://github.com/leroyvn/pinttrs/actions/workflows/tests.yml)\n[![Codecov](https://img.shields.io/codecov/c/gh/leroyvn/pinttrs?style=flat-square)](https://codecov.io/gh/leroyvn/pinttrs)\n[![Documentation Status](https://img.shields.io/readthedocs/pinttrs?style=flat-square)](https://pinttrs.readthedocs.io)\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-black?style=flat-square)](https://black.readthedocs.io)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-blue?style=flat-square&labelColor=orange)](https://pycqa.github.io/isort)\n\n## Motivation\n\nThe amazing [`attrs`](https://www.attrs.org) library is a game-changer when it\ncomes to writing classes. Its initialisation sequence notably allows for\nautomated conversion and verification of attribute values. This package is an\nattempt at designing a system to apply units automatically and reliably to\nattributes with [Pint](https://pint.readthedocs.io).\n\n## Features\n\n- Attach automatically units to unitless values passed to initialise an attribute\n- Verify unit compatibility when assigning a value to an attribute\n- Interpret units in dictionaries with a simple syntax\n- Define unit context to vary unitless value interpretation dynamically\n\nCheck the [documentation](https://pinttrs.readthedocs.io) for more detail.\n\n## License\n\nPinttrs is distributed under the terms of the\n[MIT license](https://choosealicense.com/licenses/mit/).\n\n## About\n\nPinttrs is written and maintained by [Vincent Leroy](https://github.com/leroyvn).\n\nDevelopment is supported by [Rayference](https://www.rayference.eu).\n\nPinttrs is a component of the\n[Eradiate radiative transfer model](https://www.eradiate.eu).\n\nThe Pinttrs logo is based on\n[Agus Nugroho](https://www.iconfinder.com/nugrohoagus)'s glass icon and parts of\nthe ``attrs`` logo.\n",
    'author': 'Vincent Leroy',
    'author_email': 'vincent.leroy@rayference.eu',
    'maintainer': 'Vincent Leroy',
    'maintainer_email': 'vincent.leroy@rayference.eu',
    'url': 'https://github.com/leroyvn/pinttrs',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4',
}


setup(**setup_kwargs)
